/* vim: set noet ts=4 sw=4:
 *
 * Copyright (c) 2002-2009 Martin A. Godisch <martin@godisch.de>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "config.h"
#include "constants.h"

#include <assert.h>
#include <libintl.h>
#include <ncursesw/ncurses.h>
#include <stdlib.h>
#include <string.h>

#include "charset.h"
#include "latrine.h"
#include "memory.h"

#define _(string) gettext(string)

static int
	mapcount[2] = {0, 0};
static struct {
	wchar_t from;
	char *to;
} mapping[2][MAX_KEYMAP];

wint_t digraph_table[] = {
	'A',	'\'',	193,
	'A',	':',	196,
	'a',	'\'',	225,
	'a',	':',	228,
	'C',	'<',	268,
	'c',	'<',	269,
	'D',	'<',	270,
	'd',	'<',	271,
	'E',	'\'',	201,
	'e',	'\'',	233,
	'I',	'\'',	205,
	'i',	'\'',	237,
	'L',	'\'',	313,
	'L',	'<',	317,
	'l',	'\'',	314,
	'l',	'<',	318,
	'N',	'?',	209,
	'N',	'<',	327,
	'n',	'?',	241,
	'n',	'<',	328,
	'O',	'\'',	211,
	'O',	'>',	212,
	'o',	'\'',	243,
	'o',	'>',	244,
	'R',	'\'',	340,
	'r',	'\'',	341,
	'S',	'<',	352,
	's',	'<',	353,
	'T',	'<',	356,
	't',	'<',	357,
	'U',	'\'',	218,
	'u',	'\'',	250,
	'Y',	'\'',	221,
	'y',	'\'',	253,
	'Z',	'<',	381,
	'z',	'<',	382,
	'!',	'I',	161,
	'?',	'I',	191,
	0
};

inline void
make_digraph(wint_t alpha, wint_t *special)
{
	wint_t *i;

	for (i = digraph_table; *i != 0; i += 3)
		if (alpha == i[0] && *special == i[1]) {
			*special = i[2];
			return;
		}
	fprintf(LOG, "make_digraph: unknown digraph: alpha=%u *special=%u\n", alpha, *special);
}

static inline int
quoted(const char *start, const char *pos)
{
	char *s;

	if (start == NULL || pos < start)
		return -1;
	for (s = (char*)pos - 1; *s == '\\' && s >= start; s--);
	return (pos - s + 1) & 1;
}

int
load_keymap(int i, const char *file)
{
	char buffer[BUFSIZE];
	wchar_t buf2[BUFSIZE];
	FILE *F  = NULL;
	char *s  = NULL;
	char *t  = NULL;
	int line = 0;

	assert(i == 0 || i == 1);
	if (file == NULL)
		return 0;
	if ((F = fopen(file, "r")) == NULL) {
		fprintf(LOG, "fopen: %s: %m", file);
		return -1;
	}
	mapcount[i] = 0;
	for (line = 1; fgets(buffer, BUFSIZE, F) != 0; line++) {
		if (*buffer == '\n' || *buffer == '#')
			continue;
		if ((s = index(buffer, '\n')) != NULL) /* FIXME: handle NL */
			*s = 0;
		for (t = buffer; (t = index(t, '\t')) != NULL && quoted(buffer, t); t++);
		if (t == NULL) {
			fprintf(LOG, _("%s:%d: ignoring invalid key mapping"), file, line);
			continue;
		}
		*t++ = 0;
		for (s = buffer; (s = index(s, '\\')) != NULL; s++)
			memmove(s, s + 1, strlen(s) + 1);
		for (s = t; (s = index(s, '\\')) != NULL; s++)
			memmove(s, s + 1, strlen(s) + 1);
		mbstowcs(buf2, buffer, BUFSIZE);
		mapping[i][mapcount[i]].from = *buf2;
		STRDUP(mapping[i][mapcount[i]].to, t);
		// fprintf(LOG, "adding key mapping for input %d: %X -> %s\n", i, mapping[i][mapcount[i]].from, mapping[i][mapcount[i]].to);
		if (++mapcount[i] >= MAX_KEYMAP) {
			fprintf(LOG, _("internal error: %s, please contact the author"), "load_keymap: mapcount >= MAX_KEYMAP");
			fclose(F);
			return -1;
		}
	}
	fclose(F);
	return 0;
}

inline int
wcsappend(int lng, wchar_t *dst, wchar_t c, size_t len)
{
	size_t i, n;

	if (dst == NULL)
		return -1;
	for (i = 0; i < mapcount[lng]; i++)
		if (mapping[lng][i].from == c) {
			if ((n = wcslen(dst)) + mbslen(mapping[lng][i].to) + 1 >= len)
				return -1;
			mbstowcs(dst + n, mapping[lng][i].to, len - n);
			return 0;
		}
	if ((n = wcslen(dst)) + 1 >= len)
		return -1;
	dst[n++] = c;
	dst[n] = 0L;
	return 0;
}

inline int
wcsbackspace(wchar_t *src, wchar_t *c)
{
	size_t n;

	if (src == NULL || *src == 0L) {
		if (c != NULL)
			*c = 0L;
		return -1;
	}
	n = wcslen(src) - 1;
	if (c != NULL)
		*c = src[n];
	src[n] = 0L;
	return 1;
}

inline void
wcsreverse(wchar_t *src)
{
	wchar_t *p;
	size_t i, n;

	if (src == NULL)
		return;
	n = wcslen(src);
	WMALLOC(p, n);
	for (i = 0; i < n; i++)
		p[i] = src[n-i-1];
	wmemcpy(src, p, n);
	free(p);
}
