/* vim: set noet ts=4 sw=4:
 *
 * Copyright (c) 2002-2009 Martin A. Godisch <martin@godisch.de>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "config.h"
#include "constants.h"

#include <assert.h>
#include <libintl.h>
#include <ncursesw/ncurses.h>
#include <sqlite3.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "charset.h"
#include "latrine.h"
#include "memory.h"
#include "data.h"
#include "options.h"

#define _(string) gettext(string)

static int dict_check(void);
static int dict_student(const char *name);
static int dict_languages(void);
static int create_dictionary(void);
static int get_dictcount(const char *lang[2], const int update);
static int sqlite_select_word(struct known *known);
static int sqlite_update_knowledge(const struct known *known);
static inline int sqlite_begin(void);
static inline int sqlite_commit(void);
static inline int sqlite_rollback(void);

static sqlite3 *db = NULL;
static int student = -1;
static char sqlite_errbuf[BUFLEN];

char *dictionary = NULL;
size_t randcount = DEFAULT_RANDCOUNT;
const char *dict_errmsg = sqlite_errbuf;

int
open_dictionary(void)
{
	const char *path = dictionary;
	int result;

	assert(db == NULL);
	if (strncasecmp(dictionary, "sqlite:", 7) == 0)
		path += 7;
	else if (strncasecmp(dictionary, "sqlite3:", 8) == 0)
		path += 8;
	if (sqlite3_open_v2(path, &db, SQLITE_OPEN_READWRITE, NULL) != SQLITE_OK) {
		sqlite3_close(db);
		if (sqlite3_open_v2(path, &db, SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE, NULL) != SQLITE_OK) {
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_close(db);
			db = NULL;
		}
	}
	if (db == NULL)
		return -1;
	atexit(close_dictionary);
	if ((result = dict_check()) < 0) { // !
		sqlite3_close(db);
		db = NULL;
		return -1;
	}
	if ((student = dict_student(user)) < 0) {
		sqlite3_close(db);
		db = NULL;
		return -1;
	}
	if (dict_languages() != 0) {
		sqlite3_close(db);
		db = NULL;
		return -1;
	}
	return result;
}

static int
dict_check(void)
{
	sqlite3_stmt *stmt;
	const char *tail;
	const char *version_string = NULL;
	char *e;
	int result;
	int version = -1;

	if ((result = sqlite3_prepare_v2(db, "SELECT value FROM latrine WHERE key = 'version'", -1, &stmt, &tail)) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		if (result == 1 && strcmp(dict_errmsg, "no such table: latrine") == 0)
			return create_dictionary() == 0 ? 1 : -1;
		else
			return -1;
	}
	do switch (result = sqlite3_step(stmt)) {
		case SQLITE_ROW:
			assert(version_string == NULL);
			version_string = (char*)sqlite3_column_text(stmt, 0);
			version = strtol(version_string, &e, 10);
			if (*e != 0 || version < 0)
				version = -1;
		case SQLITE_DONE:
			break;
		default:
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			return -1;
	} while (result != SQLITE_DONE);
	if (sqlite3_finalize(stmt) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		return -1;
	}
	switch (version) {
		case -1:
			snprintf(sqlite_errbuf, BUFLEN, _("invalid or missing dictionary version string"));
			return -1;
		case 10:
			return  0;
		case  9:
		case  8:
		case  7:
		case  6:
		case  5:
		case  4:
		case  3:
		case  2:
		case  1:
		case  0:
			snprintf(sqlite_errbuf, BUFLEN, _("obsolete development dictionary"));
			return -1;
		default:
			snprintf(sqlite_errbuf, BUFLEN, _("unsupported dictionary "
					"which was created with a more recent version of LaTrine"));
			return -1;
	}
}

static int
dict_student(const char *name)
{
	sqlite3_stmt *stmt;
	const char *tail;
	int result;
	int uid = -1;
	/*
	 * get student id
	 */
	if (sqlite3_prepare_v2(db, "SELECT id FROM student WHERE name = ?1", -1, &stmt, &tail) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		return -1;
	}
	if (sqlite3_bind_text(stmt, 1, name, -1, SQLITE_STATIC) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		return -1;
	}
	do switch (result = sqlite3_step(stmt)) {
		case SQLITE_ROW:
			assert(uid == -1);
			uid = sqlite3_column_int(stmt, 0);
		case SQLITE_DONE:
			break;
		default:
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			return -1;
	} while (result != SQLITE_DONE);
	if (sqlite3_finalize(stmt) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		return -1;
	}
	if (uid != -1)
		return uid;
	/*
	 * add new student
	 */
	if (sqlite3_prepare_v2(db, "INSERT INTO student (name) VALUES (?1)", -1, &stmt, &tail) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		return -1;
	}
	if (sqlite3_bind_text(stmt, 1, name, -1, SQLITE_STATIC) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		return -1;
	}
	if (sqlite3_step(stmt) != SQLITE_DONE) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		return -1;
	}
	if (sqlite3_finalize(stmt) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		return -1;
	}
	return sqlite3_last_insert_rowid(db);
}

static int
dict_languages(void)
{
	sqlite3_stmt *stmt;
	const char *tail;
	int i;
	/*
	 * insert languages, ignore errors due to duplicates
	 */
	for (i = 0; i < 2; i++) {
		if (sqlite3_prepare_v2(db, "INSERT INTO language (name) VALUES (?1)", -1, &stmt, &tail) != SQLITE_OK) {
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			return -1;
		}
		if (sqlite3_bind_text(stmt, 1, language[i], -1, SQLITE_STATIC) != SQLITE_OK) {
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			return -1;
		}
		switch (sqlite3_step(stmt)) {
			case SQLITE_CONSTRAINT:
			case SQLITE_DONE:
				sqlite3_finalize(stmt);
				continue;
			default:
				strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
				sqlite3_finalize(stmt);
				return -1;
		}
	}
	return 0;
}

static int
create_dictionary(void)
{
	const char *def[] = {
		"CREATE TABLE latrine ("
			"key TEXT NOT NULL UNIQUE PRIMARY KEY, "
			"value TEXT)",
		"INSERT INTO latrine (key, value) VALUES ('version', '" WORDLIST_VERSION_STRING "')",
	
		"CREATE TABLE student ("
			"id INTEGER NOT NULL UNIQUE PRIMARY KEY AUTOINCREMENT, "
			"name TEXT NOT NULL UNIQUE, "
			"level INTEGER NULL)",
		"CREATE TABLE language ("
			"id INTEGER NOT NULL UNIQUE PRIMARY KEY AUTOINCREMENT, "
			"name TEXT NOT NULL UNIQUE)",
		"CREATE TABLE wordlist ("
			"id INTEGER NOT NULL UNIQUE PRIMARY KEY AUTOINCREMENT, "
			"language INTEGER NOT NULL REFERENCES language (id), "
			"word TEXT NOT NULL, "
			"UNIQUE(language, word))",
		"CREATE TABLE chapter ("
			"id INTEGER NOT NULL UNIQUE PRIMARY KEY AUTOINCREMENT, "
			"name TEXT NOT NULL UNIQUE)",
		"CREATE TABLE section ("
			"id INTEGER NOT NULL UNIQUE PRIMARY KEY AUTOINCREMENT, "
			"chapter INTEGER NOT NULL REFERENCES chapter (id), "
			"name TEXT NULL, "
			"UNIQUE(chapter, name))",
		"CREATE TABLE translation ("
			"id INTEGER NOT NULL UNIQUE, " // 1:1 for the time being
			"word1 INTEGER NOT NULL REFERENCES wordlist (id), " // 1:n for the time being
			"word2 INTEGER NOT NULL REFERENCES wordlist (id), " // n:1 for the time being
			"section INTEGER NULL REFERENCES section (id), "
			"level INTEGER NULL, "
			"enabled INTEGER NOT NULL DEFAULT 1, "
			"UNIQUE(word1, word2), "
			"PRIMARY KEY (word1, word2))",
		"CREATE TABLE knowledge ("
			"student INTEGER NOT NULL REFERENCES student (id), "
			"translation INTEGER NOT NULL REFERENCES translation (id), "
			"time TIMESTAMP, "
			"history TEXT, "
			"rate INTEGER, "
			"UNIQUE (student, translation), "
			"PRIMARY KEY (student, translation))",
	
		"CREATE VIEW vocabulary AS SELECT "
			"lang1.name AS lang1, lang2.name AS lang2, "
			"chapter.name AS chapter, section.name AS section, "
			"translation.id AS translation, "
			"word1.word AS word1, word2.word AS word2, "
			"level "
			"FROM translation "
			"LEFT OUTER JOIN section ON (translation.section = section.id) "
			"LEFT OUTER JOIN chapter ON (section.chapter = chapter.id) "
			"INNER JOIN language AS lang1 ON (word1.language = lang1.id) "
			"INNER JOIN wordlist AS word1 ON (translation.word1 = word1.id) "
			"INNER JOIN language AS lang2 ON (word2.language = lang2.id) "
			"INNER JOIN wordlist AS word2 ON (translation.word2 = word2.id) "
			"WHERE enabled = 1",
	
		"CREATE INDEX wordlist_id_index ON wordlist (id)",
	/*		"CREATE INDEX wordlist_language_index ON wordlist (language)",		*/
		"CREATE INDEX translation_id_index ON translation (id)",
		"CREATE INDEX translation_word1_index ON translation (word1)",
		"CREATE INDEX translation_word2_index ON translation (word2)",
	/*		"CREATE INDEX knowledge_student_index ON knowledge (translation)",	*/
		"CREATE INDEX knowledge_translation_index ON knowledge (translation)",
		"CREATE INDEX knowledge_time_index ON knowledge (time)",
		"CREATE INDEX knowledge_rate_index ON knowledge (rate)",
	
		"CREATE TRIGGER delete_student DELETE ON student FOR EACH ROW BEGIN "
			"DELETE FROM knowledge WHERE knowledge.student = OLD.id; "
			"END",
		"CREATE TRIGGER delete_language DELETE ON language FOR EACH ROW BEGIN "
			"DELETE FROM wordlist WHERE wordlist.language = OLD.id; "
			"END",
		"CREATE TRIGGER delete_wordlist DELETE ON wordlist FOR EACH ROW BEGIN "
			"DELETE FROM translation WHERE translation.word1 = OLD.id OR translation.word2 = OLD.id; "
			"END",
		"CREATE TRIGGER delete_chapter DELETE ON chapter FOR EACH ROW BEGIN "
			"DELETE FROM section WHERE section.chapter = OLD.id; "
			"END",
		"CREATE TRIGGER delete_section DELETE ON section FOR EACH ROW BEGIN "
			"DELETE FROM translation WHERE translation.section = OLD.id; "
			"END",
		"CREATE TRIGGER delete_translation DELETE ON translation FOR EACH ROW BEGIN "
			"DELETE FROM knowledge WHERE knowledge.translation = OLD.id; "
			"END",
	
		"CREATE TRIGGER update_student UPDATE OF id ON student FOR EACH ROW BEGIN "
			"UPDATE knowledge SET student = NEW.id WHERE student = OLD.id; "
			"END",
		"CREATE TRIGGER update_language UPDATE OF id ON language FOR EACH ROW BEGIN "
			"UPDATE wordlist SET language = NEW.id WHERE language = OLD.id; "
			"END",
		"CREATE TRIGGER update_wordlist UPDATE OF id ON wordlist FOR EACH ROW BEGIN "
			"UPDATE translation SET word1 = NEW.id WHERE word1 = OLD.id; "
			"UPDATE translation SET word2 = NEW.id WHERE word2 = OLD.id; "
			"END",
		"CREATE TRIGGER update_chapter UPDATE OF id ON chapter FOR EACH ROW BEGIN "
			"UPDATE section SET chapter = NEW.id WHERE chapter = OLD.id; "
			"END",
		"CREATE TRIGGER update_section UPDATE OF id ON section FOR EACH ROW BEGIN "
			"UPDATE translation SET section = NEW.id WHERE id = OLD.id; "
			"END",
		"CREATE TRIGGER update_translation UPDATE of id ON translation FOR EACH ROW BEGIN "
			"UPDATE knowledge SET translation = NEW.id WHERE translation = OLD.id; "
			"END",
	
		"DELETE FROM sqlite_sequence",
		NULL
	};
	int i;

	if (sqlite_begin() != 0)
		return -1;
	for (i = 0; def[i] != NULL; i++)
		if (sqlite3_exec(db, def[i], NULL, NULL, NULL) != SQLITE_OK) {
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite_rollback();
			return -1;
		}
	if (sqlite_commit() != 0) {
		sqlite_rollback();
		return -1;
	}
	return 0;
}

void
close_dictionary(void)
{
	if (db != NULL) {
		sqlite3_close(db);
		db = NULL;
	}
}

static int
get_dictcount(const char *lang[2], const int update)
{
	static int n = -1;
	sqlite3_stmt *stmt;
	const char *tail;
	int i;
	int result;

	assert(db != NULL && (lang != NULL || update != 0));
	if (lang == NULL) {
		assert(n != -1);
		n += update;
		return n;
	}
	if (sqlite3_prepare_v2(db,
			"SELECT COUNT (*) FROM (SELECT DISTINCT translation FROM vocabulary "
			"WHERE lang1 = ?1 AND lang2 = ?2)", -1, &stmt, &tail) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		return -1;
	}
	for (i = 0; i < 2; i++)
		if (sqlite3_bind_text(stmt, i + 1, lang[i], -1, SQLITE_STATIC) != SQLITE_OK) {
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			return -1;
		}
	do switch (result = sqlite3_step(stmt)) {
		case SQLITE_ROW:
			n = sqlite3_column_int(stmt, 0);
		case SQLITE_DONE:
			break;
		default:
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			return -1;
	} while (result != SQLITE_DONE);
	if (sqlite3_finalize(stmt) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		return -1;
	}
	return n;
}

int
select_word(struct known *known)
{
	static int cycle = 0;
	int i;
	sqlite3_stmt *stmt;
	const char *tail;
	int num;
	int result;
	char *history;
	char *chapter;
	char *section;

	assert(db != NULL && known != NULL);
	/*
	 * clear known record
	 */
	known->key = 0;
	if (known->word[0].word != NULL) {
		free(known->word[0].word);
		bzero(&known->word[0], sizeof(struct word));
	}
	if (known->word[1].word != NULL) {
		free(known->word[1].word);
		bzero(&known->word[1], sizeof(struct word));
	}
	if (known->section != NULL)
		free(known->section);
	bzero(known->history, HISTSIZE);
	known->rate = 0;
	/*
	 * return if dictionary is empty
	 */
	if ((num = get_dictcount(language, 0)) < 0)
		return -1;
	else if (num == 0)
		return 0;
	/*
	 * select translation
	 */
	if (sqlite3_prepare_v2(db, (cycle < 7 ? ++cycle : (cycle = 0)) ?
			"SELECT translation, history, chapter, section FROM ("
				"SELECT translation, history, chapter, section FROM ("
					"SELECT translation, time, history, chapter, section FROM ("
						"SELECT DISTINCT translation, time, rate, history, chapter, section "
						"FROM vocabulary INNER JOIN knowledge USING (translation) "
						"WHERE student = ?1 AND lang1 = ?2 AND lang2 = ?3 AND rate < 3 "
						"ORDER BY rate ASC LIMIT ?5"
					") "
					"UNION ALL "
					"SELECT translation, time, history, chapter, section FROM ("
						"SELECT DISTINCT translation, time, history, chapter, section "
						"FROM vocabulary LEFT OUTER JOIN knowledge USING (translation) "
						"WHERE (student = ?1 OR student ISNULL) AND lang1 = ?2 AND lang2 = ?3 AND rate ISNULL "
					/*	"WHERE student ISNULL AND lang1 = ?2 AND lang2 = ?3 " instead possible, but:
					 *	only when rate not null in create table (otherwise tab problem, or does it set 0? */
						"ORDER BY RANDOM() LIMIT ?5"
					") "
					"UNION ALL "
					"SELECT translation, time, history, chapter, section FROM ("
						"SELECT DISTINCT translation, time, history, chapter, section "
						"FROM vocabulary INNER JOIN knowledge USING (translation) "
						"WHERE student = ?1 AND lang1 = ?2 AND lang2 = ?3 AND rate >= 3 "
						"ORDER BY rate ASC LIMIT ?5"
					") LIMIT ?5"
				") ORDER BY time DESC LIMIT -1 OFFSET ?4"
			") ORDER BY RANDOM() LIMIT 1" :
			"SELECT translation, history, chapter, section FROM ("
				"SELECT translation, history, chapter, section FROM ("
					"SELECT DISTINCT translation, time, history, chapter, section "
					"FROM vocabulary INNER JOIN knowledge USING (translation) "
					"WHERE student = ?1 AND lang1 = ?2 AND lang2 = ?3 AND NOT time ISNULL "
					"ORDER BY time DESC LIMIT -1 OFFSET ?4"
				") ORDER BY time LIMIT ?5"
			") ORDER BY RANDOM() LIMIT 1",
			-1, &stmt, &tail) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		return -1;
	}
	if (sqlite3_bind_int(stmt, 1, student) != SQLITE_OK ||
		sqlite3_bind_text(stmt, 2, language[0], -1, SQLITE_STATIC) != SQLITE_OK ||
		sqlite3_bind_text(stmt, 3, language[1], -1, SQLITE_STATIC) != SQLITE_OK ||
		sqlite3_bind_int(stmt, 4, num < 4 ? num - 1 : 3) != SQLITE_OK ||
		sqlite3_bind_int(stmt, 5, randcount) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		return -1;
	}
	known->key = 0;
	do switch (result = sqlite3_step(stmt)) {
		case SQLITE_ROW:
			assert(known->key == 0);
			known->key = sqlite3_column_int(stmt, 0);
			assert(known->key != 0);
			/* known->history contains HISTLEN + 1 bytes */
			if ((history = (char*)sqlite3_column_text(stmt, 1)) != NULL)
				strncpy(known->history, history, HISTLEN + 1);
			else
				*known->history = 0L;
			if ((chapter = (char*)sqlite3_column_text(stmt, 2)) != NULL) {
				if ((section = (char*)sqlite3_column_text(stmt, 3)) != NULL) {
					WMALLOC(known->section, strlen(chapter) + strlen(section) + 3);
					strcpy(known->section, chapter);
					strcat(known->section, ": ");
					strcat(known->section, section);
				} else {
					WMALLOC(known->section, strlen(chapter) + 1);
					strcpy(known->section, chapter);
				}
			} else
				known->section = NULL;
			break;
		case SQLITE_DONE:
			break;
		default:
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			return -1;
	} while (result != SQLITE_DONE);
	if (sqlite3_finalize(stmt) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		return -1;
	}
	if ((i = mbslen(known->history)) < HISTLEN) {
		memset(known->history + i, ' ', HISTLEN - i);
		known->history[HISTLEN] = 0;
	}
	if ((result = sqlite_select_word(known)) != 0)
		return result;
	return 1;
}

static int
sqlite_select_word(struct known *known)
{
	sqlite3_stmt *stmt;
	const char *tail;
	char *s;
	int lng, res, ret_key;

	for (lng = 0; lng < 2; lng++) {
		if (randomize)
			if (lng == 0)
				res = sqlite3_prepare_v2(db,
					"SELECT DISTINCT wordlist.id, word FROM wordlist INNER "
					"JOIN translation ON (wordlist.id = translation.word1) "
					"WHERE translation.id = ?1 ORDER BY RANDOM()",
					-1, &stmt, &tail);
			else
				res = sqlite3_prepare_v2(db,
					"SELECT DISTINCT wordlist.id, word FROM wordlist INNER "
					"JOIN translation ON (wordlist.id = translation.word2) "
					"WHERE translation.id = ?1 ORDER BY RANDOM()",
					-1, &stmt, &tail);
		else
			if (lng == 0)
				res = sqlite3_prepare_v2(db,
					"SELECT DISTINCT wordlist.id, word FROM wordlist INNER "
					"JOIN translation ON (wordlist.id = translation.word1) "
					"WHERE translation.id = ?1 ORDER BY wordlist.id",
					-1, &stmt, &tail);
			else
				res = sqlite3_prepare_v2(db,
					"SELECT DISTINCT wordlist.id, word FROM wordlist INNER "
					"JOIN translation ON (wordlist.id = translation.word2) "
					"WHERE translation.id = ?1 ORDER BY wordlist.id",
					-1, &stmt, &tail);
		if (res != SQLITE_OK) {
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			return -1;
		}
		if (sqlite3_bind_int(stmt, 1, known->key) != SQLITE_OK) {
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			return -1;
		}
		do switch (res = sqlite3_step(stmt)) {
			case SQLITE_ROW:
				ret_key = sqlite3_column_int(stmt, 0);
				known->word[lng].key = ret_key;
				s = (char*)sqlite3_column_text(stmt, 1);
				assert(s != NULL);
				WMALLOC(known->word[lng].word, mbslen(s) + 1);
				mbstowcs(known->word[lng].word, s, mbslen(s) + 1);
				break;
			case SQLITE_DONE:
				break;
			default:
				strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
				sqlite3_finalize(stmt);
				return -1;
		} while (res != SQLITE_DONE);
		if (sqlite3_finalize(stmt) != SQLITE_OK) {
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			return -1;
		}
	}
	return 0;
}

int
update_word(struct known *w, const int hit)
{
	int i;

	assert(w != NULL && w->history != NULL);
	if (hit != 0) {
		/* w->history contains HISTLEN + 1 bytes */
		memmove(w->history + 1, w->history, HISTLEN - 1);
		w->history[0] = hit > 0 ? '+' : '-';
		w->history[HISTLEN] = 0;
	}
	/* FIXME: select and update w->rate instead of recalc? */
	for (i = 0, w->rate = 0; i < HISTLEN && w->history[i] != ' '; i++)
		switch (w->history[i]) {
			case '+':
				w->rate++;
				break;
			case '-':
				w->rate--;
		}
	if (sqlite_update_knowledge(w) != 0)
		return -1;
	return 0;
}

static int
sqlite_update_knowledge(const struct known *known)
{
	sqlite3_stmt *stmt;
	const char *tail;
	int result;

	assert(db    != NULL);
	assert(known != NULL && known->key != 0 && known->history != NULL);
	if (sqlite_begin() != 0)
		return -1;
	if ((result = sqlite3_prepare_v2(db, "INSERT OR REPLACE INTO knowledge "
			"(student, translation, time, history, rate) "
			"VALUES (?1, ?2, CURRENT_TIMESTAMP, ?4, ?3)",
			-1, &stmt, &tail)) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		sqlite_rollback();
		return -1;
	}
	if (sqlite3_bind_int (stmt, 1, student) != SQLITE_OK ||
		sqlite3_bind_int (stmt, 2, known->key)  != SQLITE_OK ||
		sqlite3_bind_int (stmt, 3, known->rate) != SQLITE_OK ||
		sqlite3_bind_text(stmt, 4, known->history, -1, SQLITE_TRANSIENT) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		sqlite_rollback();
		return -1;
	}
	do switch (result = sqlite3_step(stmt)) {
		case SQLITE_DONE:
			break;
		default:
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			sqlite_rollback();
			return -1;
	} while (result != SQLITE_DONE);
	if (sqlite3_finalize(stmt) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite_rollback();
		return -1;
	}
	if (sqlite_commit() != 0) {
		sqlite_rollback();
		return -1;
	}
	return 0;
}

int
add_translation(const struct known *known)
{
	sqlite3_stmt *stmt;
	const char *tail;
	char buffer[2][BUFLEN];
	int lng, result;

	assert(db != NULL && known != NULL && known->word[0].word != NULL && known->word[1].word != NULL);

	if (sqlite_begin() != 0)
		return -1;
	/*
	 * insert words into wordlist
	 */
	if (sqlite3_prepare_v2(db, "INSERT INTO wordlist (language, word) SELECT id, ?2 FROM language WHERE name = ?1",
				-1, &stmt, &tail) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		sqlite_rollback();
		return -1;
	}
	for (lng = 0; lng < 2; lng++) {
		wcstombs(buffer[0], known->word[lng].word, BUFLEN);
		if (sqlite3_bind_text(stmt, 1, language[lng], -1, SQLITE_STATIC) != SQLITE_OK ||
			sqlite3_bind_text(stmt, 2, buffer[0], -1, SQLITE_TRANSIENT) != SQLITE_OK) {
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			sqlite_rollback();
			return -1;
		}
		do switch (result = sqlite3_step(stmt)) {
			case SQLITE_DONE:
				break;
			case SQLITE_CONSTRAINT:
				/* probably: p->word already in database */
				/* if not, we will bail out later        */
				break;
			default:
				strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
				sqlite3_finalize(stmt);
				sqlite_rollback();
				return -1;
		} while (result != SQLITE_DONE && result != SQLITE_CONSTRAINT);
		sqlite3_reset(stmt); /* ignore return value for now, since it is */
			/* an error if there was an error before (SQLITE_CONSTRAINT) */
		if ((result = sqlite3_clear_bindings(stmt)) != SQLITE_OK) {
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			sqlite_rollback();
			return -1;
		}
	}
	if (sqlite3_finalize(stmt) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite_rollback();
		return -1;
	}
	/*
	 * insert translation into db
	 */
	result = sqlite3_prepare_v2(db, "INSERT INTO translation (id, word1, word2) "
		"SELECT IFNULL(MAX(translation.id) + 1, 1), word1.id, word2.id "
		"FROM language AS lang1 INNER JOIN wordlist AS word1 ON lang1.id = word1.language "
		"INNER JOIN language AS lang2 INNER JOIN wordlist AS word2 ON lang2.id = word2.language "
		"LEFT OUTER JOIN translation "
		"WHERE lang1.name = ?1 and word1.word = ?2 and lang2.name = ?3 and word2.word = ?4",
		-1, &stmt, &tail);
	wcstombs(buffer[0], known->word[0].word, BUFLEN);
	wcstombs(buffer[1], known->word[1].word, BUFLEN);
	if (result != SQLITE_OK ||
		sqlite3_bind_text(stmt, 1, language[0], -1,  SQLITE_STATIC) != SQLITE_OK ||
		sqlite3_bind_text(stmt, 2, buffer[0], -1, SQLITE_TRANSIENT) != SQLITE_OK ||
		sqlite3_bind_text(stmt, 3, language[1], -1,  SQLITE_STATIC) != SQLITE_OK ||
		sqlite3_bind_text(stmt, 4, buffer[1], -1, SQLITE_TRANSIENT) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		sqlite_rollback();
		return -1;
	}
	do switch (result = sqlite3_step(stmt)) {
		case SQLITE_DONE:
			break;
		default:
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			sqlite_rollback();
			if (result == SQLITE_CONSTRAINT)
				return -2;
			else
				return -1;
	} while (result != SQLITE_DONE);
	if (sqlite3_finalize(stmt) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite_rollback();
		return -1;
	}
	if (sqlite_commit() != 0) {
		sqlite_rollback();
		return -1;
	}
	return 0;
}

int
delete_translation(const int id)
{
	sqlite3_stmt *stmt;
	const char *tail;
	int result;

	assert(db != NULL);
	if (sqlite_begin() != 0)
		return -1;
	if (sqlite3_prepare_v2(db, "DELETE FROM translation WHERE id = ?1", -1, &stmt, &tail) != SQLITE_OK ||
			sqlite3_bind_int(stmt, 1, id) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite3_finalize(stmt);
		sqlite_rollback();
		return -1;
	}
	do switch (result = sqlite3_step(stmt)) {
		case SQLITE_DONE:
			break;
		default:
			strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
			sqlite3_finalize(stmt);
			sqlite_rollback();
			return -1;
	} while (result != SQLITE_DONE);
	if (sqlite3_finalize(stmt) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		sqlite_rollback();
		return -1;
	}
	if (sqlite_commit() != 0) {
		sqlite_rollback();
		return -1;
	}
	return 0;
}

static inline int
sqlite_begin(void)
{
	if (sqlite3_exec(db, "BEGIN", NULL, NULL, NULL) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		return -1;
	}
	return 0;
}

static inline int
sqlite_commit(void)
{
	if (sqlite3_exec(db, "COMMIT", NULL, NULL, NULL) != SQLITE_OK) {
		strncpy(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN);
		return -1;
	}
	return 0;
}

static inline int
sqlite_rollback(void)
{
	if (sqlite3_exec(db, "ROLLBACK", NULL, NULL, NULL) != SQLITE_OK) {
		/* keep the original error message causing the rollback */
		/* snprintf(sqlite_errbuf, sqlite3_errmsg(db), BUFLEN); */
		return -1;
	}
	return 0;
}

void
wfree(struct word *w)
{
	if (w != NULL && w->word != NULL)
		free(w->word);
}

wchar_t*
word2s(struct word* w)
{
	return w != NULL ? w->word : NULL;
}

int
s2word(struct word *w, const wchar_t *s)
{
	int l;

	if (w == NULL || s == NULL)
		return -1;
	WMALLOC(w->word, l = wcslen(s) + 1);
	wcsncpy(w->word, s, l);
	w->key = 0;
	return 0;
}

int
wordcmp(struct word *w[2], const int ignore_case)
{
	return ignore_case ? wcscasecmp(w[0]->word, w[1]->word) : wcscmp(w[0]->word, w[1]->word);
}
