/* vim: set noet ts=4 sw=4:
 *
 * Copyright (c) 2002-2009 Martin A. Godisch <martin@godisch.de>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "config.h"
#include "constants.h"

#include <libintl.h>
#include <locale.h>
#include <ncursesw/ncurses.h>
#include <pwd.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "latrine.h"
#include "charset.h"
#include "memory.h"
#include "data.h"
#include "mainscr.h"
#include "options.h"
#include "popup.h"

#define _(string) gettext(string)

FILE *LOG = NULL;
int  quit = 0;
const char *user = NULL;

static void
handler(int signo)
{
	quit = 1;
}

static void
close_log(void)
{
	fclose(LOG);
}

int
main(int argc, char *argv[])
{
	static int signals[] = { SIGALRM, SIGHUP, SIGINT, SIGTERM, SIGUSR1, SIGUSR2, 0 };
	struct passwd *pwd;
	char *home, *src, *dst;
	int i;

	setlocale(LC_ALL, "");
	bindtextdomain("latrine", LOCALEDIR);
	textdomain("latrine");

	if ((pwd = getpwuid(getuid())) == NULL) {
		fprintf(stderr, _("Failure getting information for user id %d: %m\n"), getuid());
		exit(EXIT_FAILURE);
	}
	if ((user = pwd->pw_name) == NULL) {
		fprintf(stderr, _("Failure getting name for user id %d\n"), getuid());
		exit(EXIT_FAILURE);
	}
	if ((home = getenv("HOME")) == NULL)
		if ((home = pwd->pw_dir) == NULL) {
			fprintf(stderr, _("Failure getting home directory for user id %d\n"), getuid());
			exit(EXIT_FAILURE);
		}

	MALLOC(src, strlen(home) + strlen("/.latrine/latrine-0.log") + 1);
	MALLOC(dst, strlen(home) + strlen("/.latrine/latrine-0.log") + 1);

	for (i = 5; i >= 0; i--) {
		sprintf(src, "%s/.latrine/latrine-%d.log", home, i);
		sprintf(dst, "%s/.latrine/latrine-%d.log", home, i + 1);
		if (rename(src, dst) != 0) {
			fprintf(stderr, _("Renaming %s to %s failed: %m\n"), src, dst);
			exit(EXIT_FAILURE);
		}
	}
	if ((LOG = fopen(src, "w")) == NULL) {
		fprintf(stderr, _("Cannot open logfile %s: %m\n"), src);
		exit(EXIT_FAILURE);
	}
	atexit(close_log);
	setlinebuf(LOG);

	do_conf(SYSCONFDIR "/latrine.conf", 0);
	sprintf(dst, "%s/.latrine/config", home);
	do_conf(dst, 0);
	do_opts(argc, argv);
	free(src);
	free(dst);

	if (language[0] == NULL)
		STRDUP(language[0], /* TRANSLATORS: label for input line */ _("Language 1"));
	if (language[1] == NULL)
		STRDUP(language[1], /* TRANSLATORS: label for input line */ _("Language 2"));

	load_keymap(0, keymap[0]);
	load_keymap(1, keymap[1]);

	for (i = 0; signals[i] != 0; i++)
		if (signal(signals[i], handler) == SIG_ERR)
			fprintf(LOG, "Setting handler for signal %d failed: %m\n", signals[i]);

	init_screen();

	switch (open_dictionary()) {
		case 1:
			popup(/* TRANSLATORS: window headline */ _("Open dictionary"), _("New dictionary: Tables successfully created."));
		case 0:
			break;
		default:
			popup(/* TRANSLATORS: window headline */ _("Error"), _("Opening dictionary failed: %s"), dict_errmsg);
			endwin();
			exit(EXIT_FAILURE);
	}

	main_loop();

	endwin();
	exit(EXIT_SUCCESS);
}
