/* vim: set noet ts=4 sw=4:
 *
 * Copyright (c) 2002-2009 Martin A. Godisch <martin@godisch.de>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "config.h"
#include "constants.h"

#include <assert.h>
#include <libintl.h>
#include <ncursesw/panel.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <unistd.h>

#include "addscr.h"
#include "charset.h"
#include "data.h"
#include "delscr.h"
#include "latrine.h"
#include "options.h"
#include "popup.h"
#include "readline.h"

#define _(string) gettext(string)
#define FIRST_INFO_LINE   5
#define FIRST_INPUT_LINE 11
#define FIRST_EVAL_LINE  17
#define STARTINPUT (BORDER_WIDTH + lang_len + 3)
#define MAXINPUT   (sx - 2 * BORDER_WIDTH - lang_len - 3)

static WINDOW *win = NULL;
int sy, sx;
static int reverse;
static struct known p;
static char buffer[3][BUFSIZE];
static int lang_len;

static inline void
update_mode(int step)
{
	static int mode_len = 0;
	int i, x, y;

	if (mode_len == 0) {
		mode_len = mbslen(L10N_NORMAL);
		if ((i = mbslen(L10N_REVERSE)) > mode_len)
			mode_len = i;
		if ((i = mbslen(L10N_MIXED))   > mode_len)
			mode_len = i;
	}

	if (step)
		mode = (mode + 1) % 3;

	wbkgdset(win, COLOR_PAIR(COLOR_LINE));
	mvwaddstr(win, 0, sx - mode_len - 8, "[");
	switch (mode) {
		case MODE_NORMAL:
			waddstr(win, L10N_NORMAL);
			break;
		case MODE_REVERSE:
			waddstr(win, L10N_REVERSE);
			break;
		case MODE_MIXED:
			waddstr(win, L10N_MIXED);
			break;
	}
	getyx(win, y, x);
	for (i = x; i < sx - 7; i++)
		waddch(win, ' ');
	waddch(win, ']');
}

static inline void
update_dir(int step)
{
	if (step)
		direction = (direction + 1) % 4;

	wbkgdset(win, COLOR_PAIR(COLOR_LINE));
	wmove(win, 0, sx - 5);
	switch (direction) {
		case 0:
			waddstr(win, "[>>]");
			break;
		case 1:
			waddstr(win, "[<>]");
			break;
		case 2:
			waddstr(win, "[><]");
			break;
		case 3:
			waddstr(win, "[<<]");
	}
}

static inline void
update_counter(int step)
{
	static size_t
		correct  = 0,
		mistaken = 0;
	char buf[BUFSIZE];

	switch (step) {
		case +1:
			correct++;
			break;
		case -1:
			mistaken++;
	}
	snprintf(buf, BUFSIZE, "%lu/%lu (%u %%)", correct, correct + mistaken,
		(correct + mistaken > 0) ? (int)((double)(correct * 100) / (double)(correct + mistaken) + 0.5) : 0);

	wbkgdset(win, COLOR_PAIR(COLOR_BACK));
	wmove(win, FIRST_INFO_LINE, 0);
	wclrtoeol(win);
	mvwaddstr(win, FIRST_INFO_LINE, BORDER_WIDTH, _("Student: ")); waddstr(win, user);
	mvwaddstr(win, FIRST_INFO_LINE, STARTINPUT + MAXINPUT - mbslen(buf), buf);
}

static inline void
makestatuslines(void)
{
	wbkgdset(win, COLOR_PAIR(COLOR_LINE));
	wmove(win, 0, 0); wclrtoeol(win);
	mvwaddstr(win, 0, 1, PACKAGE_STRING);

	wmove(win, sy - 1, 0); wclrtoeol(win);
	wattron (win, A_BOLD); waddstr(win, " ESC ");
	wattroff(win, A_BOLD); waddstr(win, /* TRANSLATORS: button label */ _("Exit"));
	wattron (win, A_BOLD); waddstr(win, "  TAB ");
	wattroff(win, A_BOLD); waddstr(win, /* TRANSLATORS: button label */ _("Next"));
	wattron (win, A_BOLD); waddstr(win, "  F1 ");
	wattroff(win, A_BOLD); waddstr(win, /* TRANSLATORS: button label */ _("Help"));
	wattron (win, A_BOLD); waddstr(win, "  F3 ");
	wattroff(win, A_BOLD); waddstr(win, /* TRANSLATORS: button label */ _("Mode"));
	wattron (win, A_BOLD); waddstr(win, "  F4 ");
	wattroff(win, A_BOLD); waddstr(win, /* TRANSLATORS: button label */ _("Direction"));
	wattron (win, A_BOLD); waddstr(win, "  F5 ");
	wattroff(win, A_BOLD); waddstr(win, /* TRANSLATORS: button label */ _("Add"));
	wattron (win, A_BOLD); waddstr(win, "  F6 ");
	wattroff(win, A_BOLD); waddstr(win, /* TRANSLATORS: button label */ _("Delete"));

	update_mode(0);
	update_dir(0);
}

static void
make_main_panel(void)
{
	static PANEL *panel = NULL;
	int i, y, x;

	if (panel == NULL) {
		getmaxyx(stdscr, sy, sx);
		panel = new_panel(win = newwin(sy, sx, 0, 0));
		lang_len = mbslen(language[0]) > mbslen(language[1]) ? mbslen(language[0]) : mbslen(language[1]);
	} else {
		getmaxyx(stdscr, y, x);
		if (y != sy || x != sx) {
			sy = y;
			sx = x;
			wresize(win, sy, sx);
			replace_panel(panel, win);
		}
	}
	wbkgdset(win, COLOR_PAIR(COLOR_BACK));
	wclear(win);
	mvwaddstr(win, FIRST_INPUT_LINE + 0, BORDER_WIDTH, language[0]); waddch(win, ':');
	mvwaddstr(win, FIRST_INPUT_LINE + 2, BORDER_WIDTH, language[1]); waddch(win, ':');
	makestatuslines();
	update_counter(0);
	for (i = 0; i < 3; i++)
		mvwaddnstr(win, FIRST_EVAL_LINE + i, BORDER_WIDTH, buffer[i], sx - BORDER_WIDTH); /* FIXME: mvwaddnstr */
	update_panels();
}

void
init_screen(void)
{
	initscr();
	noecho();
	nonl();
	cbreak();
	if (has_colors()) {
		start_color();
		init_pair(COLOR_BACK,  COLOR_WHITE, COLOR_BLUE);
		init_pair(COLOR_LINE,  COLOR_BLACK, COLOR_WHITE);
		init_pair(COLOR_KEY,   COLOR_WHITE, COLOR_WHITE);
		init_pair(COLOR_INPUT, COLOR_WHITE, COLOR_BLACK);
	}
	*buffer[0] = 0;
	*buffer[1] = 0;
	*buffer[2] = 0;
	make_main_panel();
}

static int
check(const wchar_t *answer)
{
	struct word buffer;
	struct word *w[2];

	w[0] = &buffer;
	w[1] = &p.word[reverse ? 0 : 1];
	s2word(w[0], answer);
	return wordcmp(w, ignore_case) == 0 ? 1 : -1;
}

static int
callback(wint_t r, wint_t c, const void *udata)
{
	const struct known *word = (const struct known*)(udata);
	pid_t pid;

	switch (r) {
	case KEY_CODE_YES:
		switch (c) {
		case KEY_RESIZE:
			make_main_panel();
			/* need to get out for input refresh */
			return KEY_RESIZE;
		case KEY_F1:
			endwin();
			if ((pid = fork()) == 0)
				execlp("man", "man", "1", "latrine", NULL);
			assert(pid > 0); /* FIXME: errmsg */
			/* FIXME: close(stdin) etc.? */
			waitpid(pid, NULL, 0);
			doupdate(); /* neccessary? */
			/* no need to get out */
			break;
		case KEY_F3:
			update_mode(1);
			/* need to get out for another word */
			return KEY_F3;
		case KEY_F4:
			update_dir(1);
			/* need to get out for input refresh */
			return KEY_F4;
		case KEY_F5:
			if (add_word() >= 0)
				return KEY_F5;
			/* need to display the first word */
			/* FIXME: not neccessary if 1 before and now esc */
			break;
		case KEY_F6:
			delete_word(word);
			/* need to get out for another word */
			return KEY_TAB;
		}
		break;
	case OK:
		switch (c) {
		case KEY_TAB:
			/* need to get out for another word */
			return KEY_TAB;
		case KEY_CTRL_L:
			make_main_panel();
			/* need to get out for input refresh */
			return KEY_CTRL_L;
		}
	}
	/* default: ignore */
	return 0;
}

static int
read_input(wchar_t *input[2], const int len[2])
{
	int i = 0;

	while (1) {
		readline(win, FIRST_INPUT_LINE + (reverse ? 2 : 0), STARTINPUT, MAXINPUT, reverse ? 1 : 0, direction / 2, input[0], 0, NULL, NULL);
		switch (i = readline(win, FIRST_INPUT_LINE + (reverse ? 0 : 2), STARTINPUT, MAXINPUT, reverse ? 0 : 1, direction % 2, input[1], len[1], callback, &p)) {
		case KEY_F4:
		case KEY_CTRL_L:
		case KEY_RESIZE:
			continue;
		case KEY_ESC:
		case KEY_RETURN:
		case KEY_TAB:
		case KEY_F3:
		case KEY_F5:
			return i;
		}
	}
	return 0; /* not reached */
}

void
main_loop(void)
{
	wchar_t input[BUFSIZE];
	const int lp[2] = { BUFSIZE, BUFSIZE };
	wchar_t *wp[3];
	int status;
	int repeat = 0;

	bzero(&p, sizeof(struct known));
	wp[1] = input;
	for (;;) {
		if (!repeat) {
			/*
			 * repeat: let the not-known word be typed to remember it
			 */
			reverse = mode == MODE_REVERSE || (mode == MODE_MIXED && random() & 1);
			switch (select_word(&p)) {
				case 1:
					wp[0] = word2s(&p.word[reverse ? 1 : 0]);
					wp[2] = word2s(&p.word[reverse ? 0 : 1]);
					break;
				case 0:
					wp[0] = NULL;
					wp[2] = NULL;
					break;
				default:
					popup(_("Error"), "%s", dict_errmsg);
					endwin();
					exit(1);
			}
		}
		if (p.section != NULL) {
			wbkgdset(win, COLOR_PAIR(COLOR_BACK));
			wmove(win, FIRST_INFO_LINE + 2, 0);
			wclrtoeol(win);
			mvwaddstr(win, FIRST_INFO_LINE + 2, BORDER_WIDTH, p.section);
		}
		*input = 0L;
		switch (read_input(wp, lp)) {
		case KEY_ESC:
			return;
		case KEY_RETURN:
			if (wp[0] == NULL)
				continue;
			status = check(input);
			if (!repeat) {
				update_counter(status);
				if (update_word(&p, status) != 0)
					popup(_("Error"), _("Dictionary update failed: %s"), dict_errmsg);
			}
			snprintf(buffer[0], BUFSIZE, "%s: ", status > 0 ?
					/* TRANSLATORS: refers to a correct guess */ _("correct") :
					/* TRANSLATORS: refers to a wrong guess */ _("wrong"));
			wcstombs(buffer[0] + strlen(buffer[0]), wp[1], BUFSIZE - strlen(buffer[0]));
			wcstombs(buffer[1], wp[reverse ? 2 : 0], BUFSIZE);
			strncat(buffer[1], " -> ", BUFSIZE - strlen(buffer[1]));
			wcstombs(buffer[1] + strlen(buffer[1]), wp[reverse ? 0 : 2], BUFSIZE - strlen(buffer[1]));
			snprintf(buffer[2], BUFSIZE, "%s: |%s|",
					/* TRANSLATORS: refers to a history of previously correct/wrong guesses */ _("history"), p.history);
			wbkgdset(win, COLOR_PAIR(COLOR_BACK));
			mvwaddnstr(win, FIRST_EVAL_LINE + 0, BORDER_WIDTH, buffer[0], sx - BORDER_WIDTH); wclrtoeol(win);
			mvwaddnstr(win, FIRST_EVAL_LINE + 1, BORDER_WIDTH, buffer[1], sx - BORDER_WIDTH); wclrtoeol(win);
			mvwaddnstr(win, FIRST_EVAL_LINE + 2, BORDER_WIDTH, buffer[2], sx - BORDER_WIDTH); wclrtoeol(win);
			repeat = status == -1;
			break;
		case KEY_TAB:
			if (wp[0] != NULL && update_word(&p, 0) != 0)
				break;
			repeat = 0;
			continue;
		}
	}
}
