/* vim: set noet ts=4 sw=4:
 *
 * Copyright (c) 2002-2009 Martin A. Godisch <martin@godisch.de>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "config.h"
#include "constants.h"

#include <getopt.h>
#include <errno.h>
#include <libintl.h>
#include <ncursesw/ncurses.h>
#include <sqlite3.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "options.h"
#include "memory.h"
#include "data.h"
#include "latrine.h"

#define _(string) gettext(string)

int  digraph       = 0;
int  direction     = 0;
int  ignore_case   = 0;
int  mode          = MODE_NORMAL;
int  randomize     = 0;
char *keymap[2]    = {NULL, NULL};
const char *language[2]  = {NULL, NULL};

int
do_conf(const char *rcfile, const int raise_error)
{
	FILE *F = NULL;
	char
		buffer[BUFSIZE],
		*s      = NULL,
		*optarg = NULL;
	int line;

	if ((F = fopen(rcfile, "r")) == NULL) {
		if (errno != ENOENT || raise_error) {
			fprintf(stderr, _("Cannot open config file \"%s\": %m\n"), rcfile);
			exit(EXIT_FAILURE);
		}
		return errno;
	}
	for (line = 1; fgets(buffer, sizeof(buffer), F) != NULL; line++) {
		if ((s = index(buffer, '\n')) != NULL)
			*s = 0;
		while ((s = index(buffer, ' ')) || (s = index(buffer, '\t')))
			do {
				*s = *(s+1);
			} while (*s++);
		if (*buffer == 0 || *buffer == '#')
			continue;
		if ((optarg = index(buffer, '=')) != NULL)
			*optarg++ = 0;
		if (strcmp(buffer, "dict") == 0 && optarg)
			STRDUP(dictionary, optarg)
		else if (strcmp(buffer, "config") == 0)
			fprintf(stderr, /* TRANSLATORS: refers to options in a config file */
					_("ignored %s in %s:%d\n"), "--config", rcfile, line);
		else if (strcmp(buffer, "digraph") == 0)
			digraph = 1;
		else if ((strcmp(buffer, "direction") == 0 || strcmp(buffer, "dir") == 0) && optarg) {
			direction = strtol(optarg, &s, 10);
			if (*s != 0 || direction < 0 || direction >= 4) {
				fprintf(stderr, _("invalid direction in %s:%d -- %s\n"), rcfile, line, optarg);
				exit(EXIT_FAILURE);
			}
		} else if (strcmp(buffer, "ignore-case") == 0)
			ignore_case = 1;
		else if (strcmp(buffer, "language") == 0 && optarg) {
			if (*optarg == ':') {
				STRDUP(language[1], optarg + 1);
			} else {
				STRDUP(language[0], optarg);
				if ((s = index(language[0], ':')) != NULL) {
					*s++  = 0;
					language[1] = s;
				}
			}
		} else if (strcmp(buffer, "keymap1") == 0 && optarg) {
			FREE(keymap[0]);
			STRDUP(keymap[0], optarg);
		} else if (strcmp(buffer, "keymap2") == 0 && optarg) {
			FREE(keymap[1]);
			STRDUP(keymap[1], optarg);
		} else if (strcmp(buffer, "mode") == 0 && optarg) {
			if (strcasecmp(optarg, "normal") == 0)
				mode = MODE_NORMAL;
			else if (strcasecmp(optarg, "reverse") == 0)
				mode = MODE_REVERSE;
			else if (strcasecmp(optarg, "mixed") == 0)
				mode = MODE_MIXED;
			else {
				fprintf(stderr, _("invalid mode in %s:%d -- %s\n"), rcfile, line, optarg);
				exit(EXIT_FAILURE);
			}
		} else if (strcmp(buffer, "random") == 0 && optarg) {
			randcount = strtol(optarg, &s, 10);
			if (*s != 0 || randcount <= 0) {
				fprintf(stderr, _("invalid maximum random index in %s:%d -- %s\n"), rcfile, line, optarg);
				exit(EXIT_FAILURE);
			}
		} else if (strcmp(buffer, "randomize") == 0)
			randomize = 1;
		else if (strcmp(buffer, "help") == 0) {
			fprintf(stderr, _("ignored %s in %s:%d\n"), "--help", rcfile, line);
		} else if (strcmp(buffer, "version") == 0) {
			fprintf(stderr, _("ignored %s in %s:%d\n"), "--version", rcfile, line);
		} else {
			fprintf(stderr, _("invalid or incomplete keyword in %s:%d\n"), rcfile, line);
			exit(EXIT_FAILURE);
		}
	}
	fclose(F);
	return 0;
}

void
do_opts(int argc, char *argv[])
{
	static struct option long_opts[] = {
		{"config",      1, NULL, 'c'},
		{"digraph",     0, NULL, 'g'},
		{"direction",   1, NULL, 'd'},
		{"dir",         1, NULL,  1 },
		{"ignore-case", 0, NULL, 'i'},
		{"keymap1",     1, NULL, '1'},
		{"keymap2",     1, NULL, '2'},
		{"language",    1, NULL, 'l'},
		{"mode",        1, NULL, 'm'},
		{"random",      1, NULL, 'n'},
//		{"randomize",   0, NULL, 'r'},
		{"help",        0, NULL, 'h'},
		{"version",     0, NULL, 'v'},
		{NULL,          0, NULL,  0}};
	char *s;
	int  i;

//	while ((i = getopt_long(argc, argv, "c:d:gil:1:2:m:n:rhv", long_opts, NULL)) != -1) {
	while ((i = getopt_long(argc, argv, "c:d:gil:1:2:m:n:hv", long_opts, NULL)) != -1) {
		switch (i) {
		case 'c':
			assert(optarg);
			do_conf(optarg, 1);
			break;
		case 'd':
			assert(optarg);
			direction = strtol(optarg, &s, 10);
			if (*s != 0 || direction < 0 || direction >= 4) {
				fprintf(stderr, _("invalid direction -- %s\n"), optarg);
				exit(EXIT_FAILURE);
			}
			break;
		case 'g':
			digraph = 1;
			break;
		case 'i':
			ignore_case = 1;
			break;
		case 'l':
			assert(optarg);
			if (*optarg == ':') {
				STRDUP(language[1], optarg + 1);
			} else {
				STRDUP(language[0], optarg);
				if ((s = index(language[0], ':')) != NULL) {
					*s++  = 0;
					language[1] = s;
				}
			}
			break;
		case '1':
		case '2':
			assert(optarg);
			FREE(keymap[i-'1']); /* FIXME */
			keymap[i-'1'] = optarg;
			break;
		case 'm':
			assert(optarg);
			if (strcasecmp(optarg, "normal") == 0)
				mode = MODE_NORMAL;
			else if (strcasecmp(optarg, "reverse") == 0)
				mode = MODE_REVERSE;
			else if (strcasecmp(optarg, "mixed") == 0)
				mode = MODE_MIXED;
			else {
				fprintf(stderr, _("invalid mode -- %s\n"), optarg);
				exit(EXIT_FAILURE);
			}
			break;
		case 'n':
			assert(optarg);
			randcount = strtol(optarg, &s, 10);
			if (*s != 0 || randcount <= 0) {
				fprintf(stderr, _("invalid maximum random index -- %s\n"), optarg);
				exit(EXIT_FAILURE);
			}
			break;
//		case 'r':
//			randomize = 1;
//			break;
		case 'h':
			printf(/* TRANSLATORS: please wrap lines, which are longer than 79 characters */
				_("Usage: %s [options] [dictionary|path]\n"
				"  -d, --dir[ection]={0|1|2|3}         initial direction for input fields,\n"
				"  -c, --config=file        read config file,\n"
				"  -g, --digraph            enable input of digraphs,\n"
				"  -h, --help               display this command line summary,\n"
				"  -i, --ignore-case        compare case-insensitively,\n"
				"  -1, --keymap1=path       the input keymap for language 1,\n"
				"  -2, --keymap2=path       the input keymap for language 2,\n"
				"  -l, --language=lang1:lang2          the languages (max. %d chars each),\n"
				"  -m, --mode={normal|reverse|mixed}   specifies, in what direction to ask,\n"
				"  -n, --random=max         the maximum random index (defaults to %d),\n"
//				"  -r, --randomize          shuffle words in 1:n/n:1/n:m translations,\n"
				"  -v, --version            display the version number.\n"
				"Please read the manual page for further information.\n"),
				argv[0], MAX_LANG, DEFAULT_RANDCOUNT);
			exit(EXIT_SUCCESS);
		case 'v':
			printf(_("%s linked with %s, %s %s\n"), PACKAGE_STRING, curses_version(), "sqlite", SQLITE_VERSION);
			printf("Copyright (c) 2002-%d Martin A. Godisch <martin@godisch.de>\n", RELEASE_YEAR);
			printf(/* TRANSLATORS: please add your own credits for your translation here */
				_("English localization by Martin A. Godisch <martin@godisch.de>\n"));
			printf("\nThis program is free software: you can redistribute it and/or modify it under\n"
				"the terms of the GNU General Public License as published by the Free Software\n"
				"Foundation, either version 3 of the License, or (at your option) any later\n"
				"version.  This program is distributed in the hope that it will be useful, but\n"
				"WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or\n"
				"FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more\n"
				"details.  You should have received a copy of the GNU General Public License\n"
				"along with this program.  If not, see <http://www.gnu.org/licenses/>.\n");
			exit(EXIT_SUCCESS);
		case '?':
			exit(EXIT_FAILURE);
		}
	}
	if (argc > optind)
		dictionary = argv[optind++];
	if (argc > optind) {
		fprintf(stderr, _("invalid argument -- %s\n"), argv[optind]);
		exit(EXIT_FAILURE);
	}
	if (dictionary == NULL) {
		fprintf(stderr, _("dictionary missing\n"));
		exit(EXIT_FAILURE);
	}
}
