/* vim: set noet ts=4 sw=4:
 *
 * Copyright (c) 2002-2009 Martin A. Godisch <martin@godisch.de>.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "config.h"
#include "constants.h"

#include <assert.h>
#include <ncursesw/panel.h>
#include <stdlib.h>
#include <string.h>

#include "readline.h"
#include "charset.h"
#include "latrine.h"
#include "memory.h"
#include "options.h"

static void
update(WINDOW *win, int y, int x, int len, int r2l, wchar_t *input)
{
	wchar_t wcbuf[BUFSIZE];
	char    mbbuf[BUFSIZE];
	int     i, n;

	if ((n = wcslen(input)) < len)
		wmemcpy(wcbuf, input, n + 1);
	else {
		n = n < BUFSIZE - 1 ? len - 1 : len;
		wmemcpy(wcbuf, input + wcslen(input) - n, n + 1);
	}
	wmove(win, y, x);
	if (r2l) {
		wcsreverse(wcbuf);
		wcstombs(mbbuf, wcbuf, BUFSIZE);
		for (i = 0; i < len - n; i++)
			waddch(win, ' ');
		waddstr(win, mbbuf);
		wmove(win, y, x + len - n - (n == len ? 0 : 1));
	} else {
		wcstombs(mbbuf, wcbuf, BUFSIZE);
		waddstr(win, mbbuf);
		for (i = n; i < len; i++)
			waddch(win, ' ');
		wmove(win, y, x + n - (n == len ? 1 : 0));
	}
	update_panels();
	doupdate();
}

int
readline(WINDOW *win, int y, int x, int len, int lang, int r2l, wchar_t *result, size_t rlen,
	int (*callback)(wint_t, wint_t, const void*), const void *udata)
{
	wchar_t buffer[BUFSIZE];
	wchar_t alpha  = 0;
	int     status = 0;
	int     dig    = 0;
	wint_t  c, r;

	keypad(win, TRUE);
	wbkgdset(win, COLOR_PAIR(COLOR_INPUT));
	mvwaddch(win, y, x, ' ');
	mvwaddch(win, y, x + len - 1, ' ');
	*buffer = 0L;
	if (result != NULL)
		wcsncpy(buffer, result, BUFSIZE);
	do {
		update(win, y, x + 1, len - 2, r2l, buffer);
		if (rlen == 0)
			return 0;
		nodelay(win, 0);
		while ((r = wget_wch(win, &c)) != ERR) {
			nodelay(win, 1);
			switch (r) {
			case KEY_CODE_YES:
				switch (c) {
				case KEY_BACKSPACE:
					wcsbackspace(buffer, &alpha);
					dig = (dig == 1) ? 2 : 0;
					break;
				default:
					if (callback != NULL) {
						status = callback(r, c, udata);
						wbkgdset(win, COLOR_PAIR(COLOR_INPUT));
					}
				}
				break;
			case OK:
				switch (c) {
				case KEY_ESC:
				case KEY_RETURN:
					status = c;
					break;
				case KEY_CTRL_L:
				case KEY_TAB:
					if (callback != NULL) {
						status = callback(r, c, udata);
						wbkgdset(win, COLOR_PAIR(COLOR_INPUT));
					}
					dig = 0;
					break;
				case KEY_CTRL_U:
					*buffer = 0L;
					dig = 0;
					break;
				default:
					if (digraph && dig == 2)
						make_digraph(alpha, &c);
					if (wcsappend(lang, buffer, c, rlen) == -1) {
						wcsbackspace(buffer, NULL);
						wcsappend(lang, buffer, c, rlen);
					}
					dig = 1;
				}
			}
		}
	} while (status == 0);
	wcsncpy(result, buffer, rlen);
	return status;
}

void
read_enter(WINDOW *win)
{
	wint_t wch;

	nodelay(win, 0);
	while (wget_wch(win, &wch) == ERR || wch != KEY_RETURN);
}

/* more than one spaces are skipped on break, but kept when not */
int
make_linebreaks(char ***result, const char *line, int width)
{
	char cur_ln[BUFLEN];
	char *buffer, *p, *q, *r, *s, *t;
	int i;

	if (result == NULL)
		return -1;
	*result = NULL;
	if (line == NULL || *line == 0L)
		return 0;
	STRDUP(buffer, line);
	p = buffer;
	while ((p = index(p, '\t')) != NULL)
		*p = ' ';
	for (p = r = buffer, i = 0; r != NULL; p = q) {
		while (*p == ' ')
			p++;
		for (q = r = p, *cur_ln = 0; r != NULL && mbslen(cur_ln) <= width;) {
			q = r;
			if (*r == 0) {
				r = NULL;
				break;
			}
			if (*r == '\n')
				break;
			while (*r == ' ')
				r++;
			s = index(r, ' ');
			t = index(r, '\n');
			if (s == NULL) {
				if (t == NULL) {
					r = index(r, 0);
				} else {
					r = t;
				}
			} else if (t == NULL) {
				r = s;
			} else if (s < t) {
				r = s;
			} else {
				r = t;
			}
			memcpy(cur_ln, p, r - p);
			cur_ln[r-p] = 0;
		}
		REALLOC(*result, (i + 2) * sizeof(char*));
		STRNDUP((*result)[i++], p, q - p);
		if (*q == '\n')
			q++;
	}
	(*result)[i] = NULL;
	free(buffer);
	return i;
}

void
free_linebreaks(char **p)
{
	char **q;

	if (p == NULL)
		return;
	for (q = p; *q != NULL; q++)
		free(*q);
	free(p);
}
